/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Random utilities not classifiable elsewhere. Many are not xconq-specific. */

#include "conq.h"

#include <time.h>

/* To accommodate certain obsolete (pre-ANSI) Unix systems. */
#ifndef time_t
#define time_t long
#endif

/* Direction-to-delta conversions. */

int dirx[] = DIRX;
int diry[] = DIRY;

/* Random number handling is important to game but terrible/nonexistent
   in some systems.  Do it ourselves and hope it's better... */

/* The random state *must* be at least 32 bits. */

long initrandstate = -1;

long randstate = -1;

/* Seed can come from elsewhere, for repeatability.  Otherwise, it comes
   from the current time, scaled to a point where 32-bit arithmetic won't
   overflow. */

void
init_random(seed)
int seed;
{
    time_t tm;
    	
    if (seed > 0) {
	if (randstate >= 0) {
	    init_warning("Randstate was %d, now being set to %d",
			 randstate, seed);
	}
	randstate = seed;
    } else {
#ifdef MPW
	/* what to do here? */
#else
    	time(&tm);
#endif
    	randstate = (long) tm;
    }
    /* Whatever its source, put the randstate into known range (0 - 99999). */
    randstate = abs(randstate);
    randstate %= 100000L;
    /* This is kept around for the sake of error reporting. */
    initrandstate = randstate;
}

/* Numbers lifted from Numerical Recipes, p. 198. */
/* Arithmetic must be 32-bit. */

random(m)
int m;
{
    randstate = (8121 * randstate + 28411) % 134456L;
    return ((m * randstate) / 134456L);
}

/* Percentage probability, with bounds checking. */

probability(prob)
int prob;
{
    if (prob <= 0) return FALSE;
    if (prob >= 100) return TRUE;
    return (random(100) < prob);
}

#ifdef DEBUGGING

/* This tracks our total space allocation. */

#define NUMMALLOCRECS 100

int overflow = 0;

int totmalloc = 0;

int nextmalloc = 0;

int copymalloc = 0;

struct a_malloc {
    int size;
    int count;
} *mallocs = NULL;

/* Given an amount of memory allocated, record it with the others. */

record_malloc(amt)
int amt;
{
    int i;

    if (mallocs == NULL) {
      /* Don't use xmalloc here!! */
      mallocs = (struct a_malloc *)
	malloc(NUMMALLOCRECS * sizeof(struct a_malloc));
		overflow = 0;
	}
    for (i = 0; i < nextmalloc; ++i) {
	if (amt == mallocs[i].size) {
	    ++(mallocs[i].count);
	    return;
	}
    }
    if (nextmalloc < NUMMALLOCRECS) {
	mallocs[nextmalloc].size = amt;
	mallocs[nextmalloc].count = 1;
	++nextmalloc;
    } else {
    	/* Add to the overflow bucket. */
    	overflow += amt;
    }
}

int
malloc_record_compare(m1, m2)
struct a_malloc *m1, *m2;
{
	return (m1->size * m1->count - m2->size * m2->count);
}

/* Display memory consumption and reset records.  This does not account for freeing. */

report_malloc()
{
    extern int lispmalloc, numsymbols, maxsymbols;
    int i;

 	Dprintf("Mallocs since last report:\n");
    Dprintf("    Amount =  Bytes x  Times\n");
    /* Sort the entries. */
    qsort(mallocs, nextmalloc, sizeof(struct a_malloc), malloc_record_compare);
    for (i = 0; i < nextmalloc; ++i) {
	Dprintf("%10d = %6d x %6d\n",
		mallocs[i].size * mallocs[i].count,
		mallocs[i].size,  mallocs[i].count);
    }
    if (overflow > 0) Dprintf("%10d other\n", overflow);
    Dprintf("Total malloced = %d bytes.\n", totmalloc);
    Dprintf("String copies = %d bytes.\n", copymalloc);
    Dprintf("Lisp malloced = %d bytes.\n", lispmalloc);
    Dprintf("%d/%d symbols interned.\n", numsymbols, maxsymbols);
    nextmalloc = overflow = 0;
    totmalloc = copymalloc = lispmalloc = 0;
}

#endif /* DEBUGGING */

/* Catch any malloc failures. */

int
xmalloc(amt)
unsigned amt;
{
    int value = (int) malloc (amt);

    if (value == 0) {
	/* This is pretty serious, just have to get out quickly. */
	run_error("Memory exhausted!!");
	exit(1);  /* case run_error doesn't do it */
    }
    /* Be clean, in case callers aren't. */
    bzero(value, amt);
#ifdef DEBUGGING
    if (1 /* should be "Debug", but needs to be turned on right away */) {
	totmalloc += amt;
	record_malloc(amt);
    }
#endif
    return value;
}

/* Copy a string to newly-allocated space.  The new space is never freed. */

char *
copy_string(str)
char *str;
{
    int len = strlen(str);
    char *rslt;

    rslt = (char *) xmalloc(len + 1);
    strcpy(rslt, str);
#ifdef DEBUGGING
    copymalloc += len + 1;
#endif
    return rslt;
}

/* Get a *numeric* index into a string (more useful than ptr, in xconq). */
/* Return -1 on failed search. */

iindex(ch, str)
char ch, *str;
{
    int i;

    if (str == NULL) return (-1);
    for (i = 0; str[i] != '\0'; ++i) if (ch == str[i]) return i;
    return (-1);
}

/* This little routine goes at the end of all switch statements on internal
   data values. */

case_panic(str, var)
char *str;
int var;
{
    run_error("Panic! Unknown %s %d", str, var);
}

/* Integer square root - good enough, no float trickery or libs needed. */

/* Improved version from Bruce Fast, via Scott Herod. */

int
isqrt(i)
int i;
{
    int j, k;

    if (i > 3) {
        for (j = i, k = -1; j >>= 2; k <<= 1);
        k = (~k + i / ~k) / 2;
	k = (k + i / k) / 2;
	k = (1 + k + i / k) / 2;
	return (k + i / k) / 2;
    } else if (i > 0) {
	return 1;
    } else {
	return 0;
    }
}

#ifdef DEBUGGING

/* Debugging flags.
   These are set up so that they can be macros or variables, as desired. */

#ifndef Debug
int Debug = FALSE;
#endif
#ifndef DebugM
int DebugM = FALSE;
#endif
#ifndef DebugG
int DebugG = FALSE;
#endif

/* This is where normal debugging output goes. */

FILE *dfp;
FILE *dgfp;
FILE *dmfp;

debug_printf(str, a1, a2, a3, a4, a5, a6)
char *str;
long a1, a2, a3, a4, a5, a6;
{
    fprintf(dfp, str, a1, a2, a3, a4, a5, a6);
}

debugm_printf(str, a1, a2, a3, a4, a5, a6)
char *str;
long a1, a2, a3, a4, a5, a6;
{
    fprintf(dmfp, str, a1, a2, a3, a4, a5, a6);
}

debugg_printf(str, a1, a2, a3, a4, a5, a6)
char *str;
long a1, a2, a3, a4, a5, a6;
{
    fprintf(dgfp, str, a1, a2, a3, a4, a5, a6);
}

#else /* n DEBUGGING */

/* This exists to provide a function placeholder. */

nothing() {}

#endif /* DEBUGGING */
