/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Definitions relating to worlds and areas in them. */

typedef struct a_world {
    int circumference;     /* number of cells going around the world */
    int daylength;         /* number of turns in a day */
    int yearlength;        /* number of turns in a year */
    int firstmidwinter;
    int summernorthpoletemperature;
    int winternorthpoletemperature;
    int summernorthequatortemperature;
    int winternorthequatortemperature;
} World;

/* Theoretically, there is no maximum size to Xconq areas, but the minimum
   size is set by mystical properties, and is not negotiable. */

#define MINWIDTH 3
#define MINHEIGHT 3

/* An "area" is always basically a rectangular array of positions.  The
   hex effect is achieved interpreting neighborliness and direction
   differently, but that's all that's needed - storagewise the rect
   array is still the best choice.  All of the "layers" are dynamically
   allocated as needed, to save (considerable!) space. */

typedef struct a_area {
    int width, height;     /* size of the area */
    int maxdim;            /* max of the two dims */
    int xwrap;             /* true if x coords wrap around */
    int latitude, longitude;  /* position within whole world */
    int cellwidth;         /* distance across one cell */
    /* Pointers to the various "layers". */
    struct a_unit **units; /* pointer to units if any */
    char *terrain;         /* terrain type at this spot */
    char **auxterrain;     /* vector of extra types */
    char *peopleside;      /* the overt side alignment of the locals */
    short *features;       /* layer of ids of features */
    short *elevations;     /* layer of elevations */
    short constelev;       /* a constant elevation */
    short **materials;     /* layer of materials in each cell */
    short *temperature;    /* layer of cell temperatures */
    short consttemp;       /* a constant temperature */
    short *cloudtypes;     /* types of clouds in the layer */
    short *cloudbottoms;   /* altitudes of bottom of cloud layer */
    short *cloudheights;   /* heights of clouds in the cloud layer */
    short *winds;          /* layer of strength/dir of winds. */
    short constwinds;
    /* These layers are for temporary use in calculations. */
    short *tmp1;
    short *tmp2;
    short *tmp3;
} Area;

/* Use this macro to get a area-spanning layer of the given type. */

#define malloc_area_layer(TYPE)  \
  ((TYPE *) xmalloc(area.width * area.height * sizeof(TYPE)))

#define zero_area_layer(ADDR, TYPE)  \
  (bzero(ADDR, area.width * area.height * sizeof(TYPE)))

/* General 2D malloced area array usage.  Names from Lisp. */

#define aref(m,x,y) ((m)[(x)+area.width*(y)])

#define aset(m,x,y,v) ((m)[(x)+area.width*(y)] = (v))

#define aadd(m,x,y,v) ((m)[(x)+area.width*(y)] += (v))

/* The unit is a raw pointer - this macro is used a *lot*.  This could
   be space-optimized by using a 16-bit unit id. */

#define unit_at(x,y) (area.units[(y) * area.width + (x)])

#define set_unit_at(x,y,u) (area.units[(y) * area.width + (x)] = (u))

/* Iterate through all units in this cell (but not their occs) . */

#define for_all_stack(x,y,var)  \
  for ((var) = unit_at((x), (y)); (var) != NULL;  (var) = (var)->nexthere)

/* Test if the terrain has been allocated yet. */

#define terrain_defined() (area.terrain != NULL)

/* The terrain at each hex is just the number of the terrain type. */

#define terrain_at(x,y) ((area.terrain)[(y) * area.width + (x)])

#define set_terrain_at(x,y,t) ((area.terrain)[(y) * area.width + (x)] = (t))

/* Auxiliary terrain array of layers. */

#define any_aux_terrain_defined() (area.auxterrain != NULL)

#define aux_terrain_defined(t)  \
  (any_aux_terrain_defined() && area.auxterrain[t] != NULL)

#define aux_terrain_at(x,y,t)  \
  ((area.auxterrain[t])[(y) * area.width + (x)])

#define set_aux_terrain_at(x,y,t,v)  \
  ((area.auxterrain[t])[(y) * area.width + (x)] = (v))

/* Elevation layer. */

#define world_is_flat() (minelev == maxelev)

#define elevations_defined() (area.elevations != NULL)

#define elev_at(x,y) ((area.elevations[(y) * area.width + (x)]))

#define set_elev_at(x,y,v) ((area.elevations[(y) * area.width + (x)]) = (v))

/* Named geographical features. */

typedef struct a_geofeature {
    int type;               /* index of the general type */
    short id;               /* which one this is */
    char *name;             /* the name of the region */
    char *typename;         /* its category, such as "island" or "bay" */
    /* something for syntax? "foo bay" vs "bay of foo" */
    short superid[3];       /* indices of containing regions */
    struct a_geofeature *next; /* arranged in a linked list */
    /* cached info about a region */
    int size;
    int mindiam;
    int maxdiam;
    int relabel;
} GeoFeature;

/* Feature layer. */

#define features_defined() (area.features != NULL)

/* The "raw feature" is its "short" identifier. */

#define raw_feature_at(x,y) ((area.features)[(y) * area.width + (x)])

#define set_raw_feature_at(x,y,f)  \
  ((area.features)[(y) * area.width + (x)] = (f))

/* Population layer. */

#define people_sides_defined() (area.peopleside != NULL)

#define people_side_at(x,y) (area.peopleside[(y) * area.width + (x)])

#define set_people_side_at(x,y,s) \
  (area.peopleside[(y) * area.width + (x)] = (s))

/* A cell might be entirely uninhabited, so need an extra value to indicate. */

#define NOBODY (MAXSIDES+1)

#define populated(x,y) (people_side_at(x,y) != NOBODY)

/* Temperature layer. */

#define temperatures_defined() (area.temperature != NULL)

#define temperature_at(x,y) (area.temperature[(y) * area.width + (x)])

#define set_temperature_at(x,y,v) \
  (area.temperature[(y) * area.width + (x)] = (v))

/* Array of material layers. */

#define any_cell_materials_defined() (area.materials != NULL)

#define cell_material_defined(m) (area.materials[m] != NULL)

#define material_at(x,y,m)  \
  ((area.materials[m])[(y) * area.width + (x)])

#define set_material_at(x,y,m,v)  \
  ((area.materials[m])[(y) * area.width + (x)] = (v))

#define materials_defined() (area.materials != NULL)

/* Handlers for scratch layers. */

#define tmp1_at(x,y) (area.tmp1[(y) * area.width + (x)])

#define set_tmp1_at(x,y,v) (area.tmp1[(y) * area.width + (x)] = (v))

#define tmp2_at(x,y) (area.tmp2[(y) * area.width + (x)])

#define set_tmp2_at(x,y,v) (area.tmp2[(y) * area.width + (x)] = (v))

#define tmp3_at(x,y) (area.tmp3[(y) * area.width + (x)])

#define set_tmp3_at(x,y,v) (area.tmp3[(y) * area.width + (x)] = (v))

/* This little macro implements wraparound in the x direction. */
/* The stupid add is for the benefit of brain-damaged mod operators */
/* that don't handle negative numbers properly. */

/* (This could use some general cleanup) */

#define wrap(x) (area.xwrap ? (((x) + area.width) % area.width) : (x))

#define wrapx(x) (area.xwrap ? (((x) + area.width) % area.width) : (x))

/* Constrain y to northern and southern edges. */

#define limit(y) (max(0, min((y), (area.height-1))))

#define interior(y) (max(1, min((y), (area.height-2))))

#define xy_in_dir(x,y,d,nx,ny) \
  (nx) = wrapx((x) + dirx[d]);  (ny) = (y) + diry[dir];

/* Iteration over all valid hex positions in a area.  These should be
   used carefully, since they don't (can't) have any grouping braces
   embedded. */

#define for_all_hexes(x,y)  \
  for (x = 0; x < area.width; x++)  \
    for (y = 0; y < area.height; y++)  \
      if (in_area(x, y))

/* This doesn't generate positions along area edges.  Typically more
   useful within game. */

#define for_all_interior_hexes(x,y)  \
  for (x = 0; x < area.width; x++)  \
    for (y = 1; y < area.height - 1; y++)  \
      if (inside_area(x, y))

/* Paths and path manipulation. */

typedef struct a_waypoint {
    int x, y;
    int note;
} Waypoint;

typedef struct a_path {
    int numwps;
    Waypoint wps[100];
} Path;

/* World-related variables. */

extern World world;

extern Area area;

extern int midturnrestore;

extern int numcelltypes;
extern int numbordtypes;
extern int numconntypes;
extern int numcoattypes;

extern int minelev;
extern int maxelev;

/* World-related functions. */

GeoFeature *create_feature();
GeoFeature *find_geo_feature();
char *feature_name_at();
char *absolute_date_string();
char *greg_date_string();
